@@ -1,17 +1,26 @@
 # -*- coding: utf-8 -*-
 
 import abc
+import collections
 import enum
 import six
 import attr
 
-from cryptoparser.common.base import Serializable
-from cryptoparser.common.base import OneByteEnumComposer, OneByteEnumParsable
-from cryptoparser.common.base import Vector, VectorParsable, VectorParsableDerived
-from cryptoparser.common.base import VectorParamNumeric, VectorParamParsable
-from cryptoparser.common.algorithm import Authentication, MAC, NamedGroup
-from cryptoparser.common.base import TwoByteEnumComposer, TwoByteEnumParsable
-from cryptoparser.common.exception import NotEnoughData, InvalidValue
+from cryptoparser.common.algorithm import Authentication, Hash, NamedGroup
+from cryptoparser.common.base import (
+    OneByteEnumComposer,
+    OneByteEnumParsable,
+    Serializable,
+    TwoByteEnumComposer,
+    TwoByteEnumParsable,
+    VariantParsable,
+    Vector,
+    VectorParamNumeric,
+    VectorParamParsable,
+    VectorParsable,
+    VectorParsableDerived,
+)
+from cryptoparser.common.exception import NotEnoughData, InvalidType
 from cryptoparser.common.parse import ParsableBase, ParserBinary, ComposerBinary
 from cryptoparser.tls.grease import TlsInvalidTypeOneByte, TlsInvalidTypeTwoByte
 from cryptoparser.tls.version import TlsProtocolVersionBase
@@ -205,11 +214,11 @@ class TlsExtensionType(Serializable, TwoByteEnumComposer, enum.Enum):
     )
 
 
-class TlsExtensions(VectorParsableDerived):
+class TlsExtensions(VectorParsable):
     @classmethod
     def get_param(cls):
         return VectorParamParsable(
-            item_class=TlsExtensionParsed,
+            item_class=TlsExtensionVariant,
             fallback_class=TlsExtensionUnparsed,
             min_byte_num=0, max_byte_num=2 ** 16 - 1
         )
@@ -311,7 +320,7 @@ class TlsExtensionParsed(TlsExtensionBase):
         parser = super(TlsExtensionParsed, cls)._check_header(parsable)
 
         if parser['extension_type'] != cls.get_extension_type():
-            raise InvalidValue(parser['extension_type'], TlsExtensionParsed, 'extension type')
+            raise InvalidType()
 
         return parser
 
@@ -380,9 +389,9 @@ class TlsECPointFormatFactory(OneByteEnumParsable):
 
 
 class TlsECPointFormat(Serializable, OneByteEnumComposer, enum.Enum):
-    UNCOMPRESSED = TlsECPointFormatParams(code=0x0)
-    ANSIX962_COMPRESSED_PRIME = TlsECPointFormatParams(code=0x1)
-    ANSIX962_COMPRESSED_CHAR2 = TlsECPointFormatParams(code=0x2)
+    UNCOMPRESSED = TlsECPointFormatParams(code=0x00)
+    ANSIX962_COMPRESSED_PRIME = TlsECPointFormatParams(code=0x01)
+    ANSIX962_COMPRESSED_CHAR2 = TlsECPointFormatParams(code=0x02)
 
 
 class TlsECPointFormatVector(VectorParsable):
@@ -398,7 +407,10 @@ class TlsECPointFormatVector(VectorParsable):
 
 @attr.s
 class TlsExtensionECPointFormats(TlsExtensionParsed):
-    point_formats = attr.ib(validator=attr.validators.instance_of(TlsECPointFormatVector))
+    point_formats = attr.ib(
+        converter=TlsECPointFormatVector,
+        validator=attr.validators.instance_of(TlsECPointFormatVector)
+    )
 
     @classmethod
     def get_extension_type(cls):
@@ -432,7 +444,7 @@ class TlsNamedCurveFactory(TwoByteEnumParsable):
         raise NotImplementedError()
 
 
-class TlsNamedCurve(TwoByteEnumComposer, enum.Enum):
+class TlsNamedCurve(TwoByteEnumComposer):
     SECT163K1 = TlsNamedCurveParams(
         code=0x0001,
         named_group=NamedGroup.SECT163K1,
@@ -507,7 +519,7 @@ class TlsNamedCurve(TwoByteEnumComposer, enum.Enum):
     )
     SECP192R1 = TlsNamedCurveParams(
         code=0x0013,
-        named_group=NamedGroup.SECP192R1,
+        named_group=NamedGroup.PRIME192V1,
     )
     SECP224K1 = TlsNamedCurveParams(
         code=0x0014,
@@ -523,7 +535,7 @@ class TlsNamedCurve(TwoByteEnumComposer, enum.Enum):
     )
     SECP256R1 = TlsNamedCurveParams(
         code=0x0017,
-        named_group=NamedGroup.SECP256R1,
+        named_group=NamedGroup.PRIME256V1,
     )
     SECP384R1 = TlsNamedCurveParams(
         code=0x0018,
@@ -533,6 +545,34 @@ class TlsNamedCurve(TwoByteEnumComposer, enum.Enum):
         code=0x0019,
         named_group=NamedGroup.SECP521R1,
     )
+    GC256A = TlsNamedCurveParams(
+        code=0x0022,
+        named_group=NamedGroup.GC256A,
+    )
+    GC256B = TlsNamedCurveParams(
+        code=0x0023,
+        named_group=NamedGroup.GC256B,
+    )
+    GC256C = TlsNamedCurveParams(
+        code=0x0024,
+        named_group=NamedGroup.GC256C,
+    )
+    GC256D = TlsNamedCurveParams(
+        code=0x0025,
+        named_group=NamedGroup.GC256D,
+    )
+    GC512A = TlsNamedCurveParams(
+        code=0x0026,
+        named_group=NamedGroup.GC512A,
+    )
+    GC512B = TlsNamedCurveParams(
+        code=0x0027,
+        named_group=NamedGroup.GC512B,
+    )
+    GC512C = TlsNamedCurveParams(
+        code=0x0028,
+        named_group=NamedGroup.GC512C,
+    )
 
     BRAINPOOLP256R1 = TlsNamedCurveParams(
         code=0x001a,
@@ -598,7 +638,10 @@ class TlsEllipticCurveVector(VectorParsable):
 
 @attr.s
 class TlsExtensionEllipticCurves(TlsExtensionParsed):
-    elliptic_curves = attr.ib(validator=attr.validators.instance_of(TlsEllipticCurveVector))
+    elliptic_curves = attr.ib(
+        converter=TlsEllipticCurveVector,
+        validator=attr.validators.instance_of(TlsEllipticCurveVector)
+    )
 
     @classmethod
     def get_extension_type(cls):
@@ -671,11 +714,11 @@ class TlsSignatureAndHashAlgorithmFactory(TwoByteEnumParsable):
 @attr.s(frozen=True)
 class HashAndSignatureAlgorithmParam(object):
     code = attr.ib(validator=attr.validators.instance_of(int))
-    hash_algorithm = attr.ib(validator=attr.validators.optional(attr.validators.in_(MAC)))
+    hash_algorithm = attr.ib(validator=attr.validators.optional(attr.validators.in_(Hash)))
     signature_algorithm = attr.ib(validator=attr.validators.optional(attr.validators.in_(Authentication)))
 
 
-class TlsSignatureAndHashAlgorithm(TwoByteEnumComposer, enum.Enum):
+class TlsSignatureAndHashAlgorithm(TwoByteEnumComposer):
     ANONYMOUS_NONE = HashAndSignatureAlgorithmParam(
         code=0x0000,
         signature_algorithm=Authentication.anon,
@@ -684,32 +727,32 @@ class TlsSignatureAndHashAlgorithm(TwoByteEnumComposer, enum.Enum):
     ANONYMOUS_MD5 = HashAndSignatureAlgorithmParam(
         code=0x0100,
         signature_algorithm=Authentication.anon,
-        hash_algorithm=MAC.MD5
+        hash_algorithm=Hash.MD5
     )
     ANONYMOUS_SHA1 = HashAndSignatureAlgorithmParam(
         code=0x0200,
         signature_algorithm=Authentication.anon,
-        hash_algorithm=MAC.SHA1
+        hash_algorithm=Hash.SHA1
     )
     ANONYMOUS_SHA224 = HashAndSignatureAlgorithmParam(
         code=0x0300,
         signature_algorithm=Authentication.anon,
-        hash_algorithm=MAC.SHA224
+        hash_algorithm=Hash.SHA2_224
     )
     ANONYMOUS_SHA256 = HashAndSignatureAlgorithmParam(
         code=0x0400,
         signature_algorithm=Authentication.anon,
-        hash_algorithm=MAC.SHA256
+        hash_algorithm=Hash.SHA2_256
     )
     ANONYMOUS_SHA384 = HashAndSignatureAlgorithmParam(
         code=0x0500,
         signature_algorithm=Authentication.anon,
-        hash_algorithm=MAC.SHA384
+        hash_algorithm=Hash.SHA2_384
     )
     ANONYMOUS_SHA512 = HashAndSignatureAlgorithmParam(
         code=0x0006,
         signature_algorithm=Authentication.anon,
-        hash_algorithm=MAC.SHA512
+        hash_algorithm=Hash.SHA2_512
     )
     RSA_NONE = HashAndSignatureAlgorithmParam(
         code=0x0001,
@@ -719,32 +762,32 @@ class TlsSignatureAndHashAlgorithm(TwoByteEnumComposer, enum.Enum):
     RSA_MD5 = HashAndSignatureAlgorithmParam(
         code=0x0101,
         signature_algorithm=Authentication.RSA,
-        hash_algorithm=MAC.MD5
+        hash_algorithm=Hash.MD5
     )
     RSA_SHA1 = HashAndSignatureAlgorithmParam(
         code=0x0201,
         signature_algorithm=Authentication.RSA,
-        hash_algorithm=MAC.SHA1
+        hash_algorithm=Hash.SHA1
     )
     RSA_SHA224 = HashAndSignatureAlgorithmParam(
         code=0x0301,
         signature_algorithm=Authentication.RSA,
-        hash_algorithm=MAC.SHA224
+        hash_algorithm=Hash.SHA2_224
     )
     RSA_SHA256 = HashAndSignatureAlgorithmParam(
         code=0x0401,
         signature_algorithm=Authentication.RSA,
-        hash_algorithm=MAC.SHA256
+        hash_algorithm=Hash.SHA2_256
     )
     RSA_SHA384 = HashAndSignatureAlgorithmParam(
         code=0x0501,
         signature_algorithm=Authentication.RSA,
-        hash_algorithm=MAC.SHA384
+        hash_algorithm=Hash.SHA2_384
     )
     RSA_SHA512 = HashAndSignatureAlgorithmParam(
         code=0x0601,
         signature_algorithm=Authentication.RSA,
-        hash_algorithm=MAC.SHA512
+        hash_algorithm=Hash.SHA2_512
     )
     DSA_NONE = HashAndSignatureAlgorithmParam(
         code=0x0002,
@@ -754,32 +797,32 @@ class TlsSignatureAndHashAlgorithm(TwoByteEnumComposer, enum.Enum):
     DSA_MD5 = HashAndSignatureAlgorithmParam(
         code=0x0102,
         signature_algorithm=Authentication.DSS,
-        hash_algorithm=MAC.MD5
+        hash_algorithm=Hash.MD5
     )
     DSA_SHA1 = HashAndSignatureAlgorithmParam(
         code=0x0202,
         signature_algorithm=Authentication.DSS,
-        hash_algorithm=MAC.SHA1
+        hash_algorithm=Hash.SHA1
     )
     DSA_SHA224 = HashAndSignatureAlgorithmParam(
         code=0x0302,
         signature_algorithm=Authentication.DSS,
-        hash_algorithm=MAC.SHA224
+        hash_algorithm=Hash.SHA2_224
     )
     DSA_SHA256 = HashAndSignatureAlgorithmParam(
         code=0x0402,
         signature_algorithm=Authentication.DSS,
-        hash_algorithm=MAC.SHA256
+        hash_algorithm=Hash.SHA2_256
     )
     DSA_SHA384 = HashAndSignatureAlgorithmParam(
         code=0x0502,
         signature_algorithm=Authentication.DSS,
-        hash_algorithm=MAC.SHA384
+        hash_algorithm=Hash.SHA2_384
     )
     DSA_SHA512 = HashAndSignatureAlgorithmParam(
         code=0x0602,
         signature_algorithm=Authentication.DSS,
-        hash_algorithm=MAC.SHA512
+        hash_algorithm=Hash.SHA2_512
     )
     ECDSA_NONE = HashAndSignatureAlgorithmParam(
         code=0x0003,
@@ -789,32 +832,100 @@ class TlsSignatureAndHashAlgorithm(TwoByteEnumComposer, enum.Enum):
     ECDSA_MD5 = HashAndSignatureAlgorithmParam(
         code=0x0103,
         signature_algorithm=Authentication.ECDSA,
-        hash_algorithm=MAC.MD5
+        hash_algorithm=Hash.MD5
     )
     ECDSA_SHA1 = HashAndSignatureAlgorithmParam(
         code=0x0203,
         signature_algorithm=Authentication.ECDSA,
-        hash_algorithm=MAC.SHA1
+        hash_algorithm=Hash.SHA1
     )
     ECDSA_SHA224 = HashAndSignatureAlgorithmParam(
         code=0x0303,
         signature_algorithm=Authentication.ECDSA,
-        hash_algorithm=MAC.SHA224
+        hash_algorithm=Hash.SHA2_224
     )
     ECDSA_SHA256 = HashAndSignatureAlgorithmParam(
         code=0x0403,
         signature_algorithm=Authentication.ECDSA,
-        hash_algorithm=MAC.SHA256
+        hash_algorithm=Hash.SHA2_256
     )
     ECDSA_SHA384 = HashAndSignatureAlgorithmParam(
         code=0x0503,
         signature_algorithm=Authentication.ECDSA,
-        hash_algorithm=MAC.SHA384
+        hash_algorithm=Hash.SHA2_384
     )
     ECDSA_SHA512 = HashAndSignatureAlgorithmParam(
         code=0x0603,
         signature_algorithm=Authentication.ECDSA,
-        hash_algorithm=MAC.SHA512
+        hash_algorithm=Hash.SHA2_512
+    )
+    GOST_R3410_01 = HashAndSignatureAlgorithmParam(
+        code=0x00ed,
+        signature_algorithm=Authentication.GOST_R3410_01,
+        hash_algorithm=Hash.GOST_R3411_94,
+    )
+    OLD_GOST_R3410_12_256 = HashAndSignatureAlgorithmParam(
+        code=0x00ee,
+        signature_algorithm=Authentication.GOST_R3410_12_256,
+        hash_algorithm=Hash.GOST_R3411_12_256,
+    )
+    OLD_GOST_R3410_12_512 = HashAndSignatureAlgorithmParam(
+        code=0x00ef,
+        signature_algorithm=Authentication.GOST_R3410_12_512,
+        hash_algorithm=Hash.GOST_R3411_12_512,
+    )
+    GOST_R3410_12_256 = HashAndSignatureAlgorithmParam(
+        code=0x4008,
+        signature_algorithm=Authentication.GOST_R3410_12_256,
+        hash_algorithm=Hash.GOST_R3411_12_256,
+    )
+    GOST_R3410_12_512 = HashAndSignatureAlgorithmParam(
+        code=0x4108,
+        signature_algorithm=Authentication.GOST_R3410_12_512,
+        hash_algorithm=Hash.GOST_R3411_12_512,
+    )
+
+    RSA_PSS_RSAE_SHA256 = HashAndSignatureAlgorithmParam(
+        code=0x0804,
+        signature_algorithm=Authentication.RSA,
+        hash_algorithm=Hash.SHA2_256
+    )
+    RSA_PSS_RSAE_SHA384 = HashAndSignatureAlgorithmParam(
+        code=0x0805,
+        signature_algorithm=Authentication.RSA,
+        hash_algorithm=Hash.SHA2_384
+    )
+    RSA_PSS_RSAE_SHA512 = HashAndSignatureAlgorithmParam(
+        code=0x0806,
+        signature_algorithm=Authentication.RSA,
+        hash_algorithm=Hash.SHA2_512
+    )
+
+    ED25519 = HashAndSignatureAlgorithmParam(
+        code=0x0807,
+        signature_algorithm=Authentication.EDDSA,
+        hash_algorithm=Hash.ED25519PH
+    )
+    ED448 = HashAndSignatureAlgorithmParam(
+        code=0x0808,
+        signature_algorithm=Authentication.EDDSA,
+        hash_algorithm=Hash.ED448PH
+    )
+
+    RSA_PSS_PSS_SHA256 = HashAndSignatureAlgorithmParam(
+        code=0x0809,
+        signature_algorithm=Authentication.RSA,
+        hash_algorithm=Hash.SHA2_256
+    )
+    RSA_PSS_PSS_SHA384 = HashAndSignatureAlgorithmParam(
+        code=0x080a,
+        signature_algorithm=Authentication.RSA,
+        hash_algorithm=Hash.SHA2_384
+    )
+    RSA_PSS_PSS_SHA512 = HashAndSignatureAlgorithmParam(
+        code=0x080b,
+        signature_algorithm=Authentication.RSA,
+        hash_algorithm=Hash.SHA2_512
     )
 
 
@@ -830,7 +941,10 @@ class TlsSignatureAndHashAlgorithmVector(VectorParsable):
 
 @attr.s
 class TlsExtensionSignatureAlgorithms(TlsExtensionParsed):
-    hash_and_signature_algorithms = attr.ib(validator=attr.validators.instance_of(TlsSignatureAndHashAlgorithmVector))
+    hash_and_signature_algorithms = attr.ib(
+        converter=TlsSignatureAndHashAlgorithmVector,
+        validator=attr.validators.instance_of(TlsSignatureAndHashAlgorithmVector)
+    )
 
     @classmethod
     def get_extension_type(cls):
@@ -852,3 +966,27 @@ class TlsExtensionSignatureAlgorithms(TlsExtensionParsed):
         header_bytes = self._compose_header(payload_composer.composed_length)
 
         return header_bytes + payload_composer.composed_bytes
+
+
+class TlsExtensionVariant(VariantParsable):
+    _VARIANTS = collections.OrderedDict(
+        [
+            (TlsExtensionType.SERVER_NAME, (TlsExtensionServerName, )),
+            (TlsExtensionType.SUPPORTED_GROUPS, (TlsExtensionEllipticCurves, )),
+            (TlsExtensionType.EC_POINT_FORMATS, (TlsExtensionECPointFormats, )),
+            (TlsExtensionType.SIGNATURE_ALGORITHMS, (TlsExtensionSignatureAlgorithms, )),
+            (TlsExtensionType.SUPPORTED_VERSIONS, (TlsExtensionSupportedVersions, )),
+        ]
+    )
+
+    @classmethod
+    def _get_variants(cls):
+        variants = collections.OrderedDict(cls._VARIANTS)
+
+        variants.update([
+            (extension_type, (TlsExtensionUnparsed, ))
+            for extension_type in TlsExtensionType
+            if extension_type not in cls._VARIANTS
+        ])
+
+        return variants
