@@ -1,17 +1,33 @@
 # -*- coding: utf-8 -*-
+# pylint: disable=too-many-lines
 
 import abc
 import calendar
+import collections
 import datetime
 import enum
 import random
 import attr
 
-from cryptoparser.common.base import Opaque, Vector, VectorParamNumeric, VectorParamParsable, VectorParsable
+import six
+
+from cryptoparser.common.base import (
+    Opaque,
+    OpaqueParam,
+    VariantParsable,
+    Vector,
+    VectorParamNumeric,
+    VectorParamParsable,
+    VectorParsable,
+)
 from cryptoparser.common.exception import NotEnoughData, InvalidValue, InvalidType
 from cryptoparser.common.parse import ParsableBase, ParserBinary, ComposerBinary
 
-from cryptoparser.tls.extension import TlsExtensions, TlsExtensionType
+from cryptoparser.tls.extension import (
+    TlsExtensionType,
+    TlsExtensions,
+    TlsSignatureAndHashAlgorithmVector,
+)
 from cryptoparser.tls.grease import TlsInvalidType, TlsInvalidTypeOneByte, TlsInvalidTypeTwoByte
 from cryptoparser.tls.version import (
     SslProtocolVersion,
@@ -55,41 +71,12 @@ class SubprotocolParser(object):
         subprotocol_parsers = self._get_subprotocol_parsers()
 
         if self._subprotocol_type in subprotocol_parsers:
-            parsed_object, unparsed_bytes = subprotocol_parsers[self._subprotocol_type].parse_immutable(parsable)
-            return parsed_object, len(parsable) - len(unparsed_bytes)
+            parsed_object, parsed_length = subprotocol_parsers[self._subprotocol_type].parse_immutable(parsable)
+            return parsed_object, parsed_length
 
         raise InvalidValue(self._subprotocol_type, TlsSubprotocolMessageBase)
 
 
-@attr.s
-class VariantParsable(ParsableBase):
-    variant = attr.ib(validator=attr.validators.instance_of(ParsableBase))
-
-    @classmethod
-    @abc.abstractmethod
-    def _get_variants(cls):
-        raise NotImplementedError()
-
-    @classmethod
-    def register_variant_parser(cls, variant_tag, parsable_class):
-        variants = cls._get_variants()
-        variants[variant_tag] = parsable_class
-
-    @classmethod
-    def _parse(cls, parsable):
-        for variant_parser in cls._get_variants().values():
-            try:
-                parsed_object, unparsed_bytes = variant_parser.parse_immutable(parsable)
-                return parsed_object, len(parsable) - len(unparsed_bytes)
-            except InvalidType:
-                continue
-
-        raise InvalidValue(parsable, cls)
-
-    def compose(self):
-        return self.variant.compose()
-
-
 class TlsSubprotocolMessageBase(ParsableBase):
     @classmethod
     @abc.abstractmethod
@@ -169,15 +156,15 @@ class TlsAlertMessage(TlsSubprotocolMessageBase):
     def _validator_level(self, attribute, value):  # pylint: disable=unused-argument
         try:
             self.level = TlsAlertLevel(value)
-        except ValueError:
-            raise InvalidValue(value, TlsAlertLevel, 'level')
+        except ValueError as e:
+            six.raise_from(InvalidValue(value, TlsAlertLevel, 'level'), e)
 
     @description.validator
     def _validator_description(self, attribute, value):  # pylint: disable=unused-argument
         try:
             self.description = TlsAlertDescription(value)
-        except ValueError:
-            raise InvalidValue(value, TlsAlertDescription)
+        except ValueError as e:
+            six.raise_from(InvalidValue(value, TlsAlertDescription), e)
 
 
 TlsSubprotocolMessageBase.register(TlsAlertMessage)
@@ -283,7 +270,7 @@ class TlsHandshakeMessage(TlsSubprotocolMessageBase):
         try:
             parser.parse_bytes('payload', parser['handshake_length'])
         except NotEnoughData as e:
-            raise NotEnoughData(e.bytes_needed)
+            six.raise_from(NotEnoughData(e.bytes_needed), e)
 
         return parser
 
@@ -296,10 +283,23 @@ class TlsHandshakeMessage(TlsSubprotocolMessageBase):
         return composer.composed_bytes
 
 
-class TlsHandshakeHelloRandomBytes(Opaque):
+class TlsHandshakeHelloRandomBytes(Vector):
     @classmethod
-    def get_byte_num(cls):
-        return 28
+    def _parse(cls, parsable):
+        composer = ComposerBinary()
+        vector_param = cls.get_param()
+        composer.compose_numeric(vector_param.min_byte_num, vector_param.item_num_size)
+
+        vector, parsed_length = super(TlsHandshakeHelloRandomBytes, cls)._parse(composer.composed_bytes + parsable)
+
+        return cls(vector), parsed_length - vector_param.item_num_size
+
+    def compose(self):
+        return super(TlsHandshakeHelloRandomBytes, self).compose()[self.get_param().item_num_size:]
+
+    @classmethod
+    def get_param(cls):
+        return OpaqueParam(min_byte_num=28, max_byte_num=28)
 
 
 @attr.s
@@ -421,7 +421,10 @@ class TlsSessionIdVector(Vector):
 
 @attr.s  # pylint: disable=too-many-instance-attributes
 class TlsHandshakeClientHello(TlsHandshakeHello):
-    cipher_suites = attr.ib(converter=TlsCipherSuiteVector)
+    cipher_suites = attr.ib(
+        converter=TlsCipherSuiteVector,
+        validator=attr.validators.instance_of(TlsCipherSuiteVector)
+    )
     protocol_version = attr.ib(
         default=TlsProtocolVersionFinal(TlsVersion.TLS1_2),
         validator=attr.validators.instance_of((TlsProtocolVersionBase, SslProtocolVersion)),
@@ -438,9 +441,13 @@ class TlsHandshakeClientHello(TlsHandshakeHello):
         default=TlsCompressionMethodVector([TlsCompressionMethod.NULL, ]),
         validator=attr.validators.instance_of(TlsCompressionMethodVector),
     )
-    extensions = attr.ib(default=TlsExtensions(()), validator=attr.validators.instance_of(TlsExtensions))
+    extensions = attr.ib(
+        default=TlsExtensions(()),
+        converter=TlsExtensions,
+        validator=attr.validators.instance_of(TlsExtensions)
+    )
     fallback_scsv = attr.ib(default=False, validator=attr.validators.instance_of(bool))
-    empty_renegotiation_info_scsv = attr.ib(default=False, validator=attr.validators.instance_of(bool))
+    empty_renegotiation_info_scsv = attr.ib(default=True, validator=attr.validators.instance_of(bool))
 
     @classmethod
     def get_handshake_type(cls):
@@ -559,7 +566,11 @@ class TlsHandshakeServerHello(TlsHandshakeHello):
         validator=attr.validators.in_(TlsCompressionMethod),
     )
     cipher_suite = attr.ib(default=None, validator=attr.validators.in_(TlsCipherSuite))
-    extensions = attr.ib(default=None, validator=attr.validators.instance_of(TlsExtensions))
+    extensions = attr.ib(
+        default=TlsExtensions([]),
+        converter=TlsExtensions,
+        validator=attr.validators.instance_of(TlsExtensions)
+    )
 
     @classmethod
     def get_handshake_type(cls):
@@ -706,6 +717,102 @@ class TlsHandshakeServerKeyExchange(TlsHandshakeMessage):
         return self._compose_header(len(self.param_bytes)) + bytes(self.param_bytes)
 
 
+class TlsClientCertificateType(enum.IntEnum):
+    RSA_SIGN = 0x01
+    DSS_SIGN = 0x02
+    RSA_FIXED_DH = 0x03
+    DSS_FIXED_DH = 0x04
+    ECDSA_SIGN = 0x40
+    RSA_FIXED_ECDH = 0x41
+    ECDSA_FIXED_ECDH = 0x42
+    GOST_SIGN256 = 0x43
+    GOST_SIGN512 = 0x44
+
+
+class TlsClientCertificateTypeVector(Vector):
+    @classmethod
+    def get_param(cls):
+        return VectorParamNumeric(
+            item_size=1,
+            min_byte_num=1,
+            max_byte_num=2 ** 8 - 1,
+            numeric_class=TlsClientCertificateType
+        )
+
+
+class TlsDistinguishedName(Opaque):
+    @classmethod
+    def get_param(cls):
+        return OpaqueParam(min_byte_num=1, max_byte_num=2 ** 16 - 1)
+
+
+class TlsDistinguishedNameVector(VectorParsable):
+    @classmethod
+    def get_param(cls):
+        return VectorParamParsable(
+            item_class=TlsDistinguishedName,
+            fallback_class=None,
+            min_byte_num=0, max_byte_num=2 ** 16 - 1
+        )
+
+
+@attr.s
+class TlsHandshakeCertificateRequest(TlsHandshakeMessage):
+    certificate_types = attr.ib(converter=TlsClientCertificateTypeVector)
+    certificate_authorities = attr.ib(converter=TlsDistinguishedNameVector)
+    supported_signature_algorithms = attr.ib(default=None)
+
+    def __attrs_post_init__(self):
+        if self.supported_signature_algorithms is not None:
+            self.supported_signature_algorithms = TlsSignatureAndHashAlgorithmVector(
+                self.supported_signature_algorithms
+            )
+
+        attr.validate(self)
+
+    @classmethod
+    def get_handshake_type(cls):
+        return TlsHandshakeType.CERTIFICATE_REQUEST
+
+    @classmethod
+    def _parse(cls, parsable):
+        handshake_header_parser = cls._parse_handshake_header(parsable)
+
+        parser = ParserBinary(handshake_header_parser['payload'])
+
+        parser.parse_parsable('certificate_types', TlsClientCertificateTypeVector)
+
+        parser_remaining = ParserBinary(parser.unparsed)
+        parser_remaining.parse_numeric('vector_length', 2)
+        if parser_remaining['vector_length'] + 2 == parser.unparsed_length:
+            supported_signature_algorithms = None
+        else:
+            parser.parse_parsable('supported_signature_algorithms', TlsSignatureAndHashAlgorithmVector)
+            supported_signature_algorithms = parser['supported_signature_algorithms']
+
+        parser.parse_parsable('certificate_authorities', TlsDistinguishedNameVector)
+
+        msg = TlsHandshakeCertificateRequest(
+            certificate_types=parser['certificate_types'],
+            certificate_authorities=parser['certificate_authorities'],
+            supported_signature_algorithms=supported_signature_algorithms,
+        )
+
+        return msg, handshake_header_parser.parsed_length
+
+    def compose(self):
+        payload_composer = ComposerBinary()
+
+        payload_composer.compose_parsable(self.certificate_types)
+        if self.supported_signature_algorithms is not None:
+            payload_composer.compose_parsable(self.supported_signature_algorithms)
+        payload_composer.compose_parsable(self.certificate_authorities)
+
+        header_bytes = self._compose_header(payload_composer.composed_length)
+
+        return header_bytes + payload_composer.composed_bytes
+
+
 class SslMessageBase(ParsableBase):
     @classmethod
     def get_message_type(cls):
@@ -879,13 +986,14 @@ class SslHandshakeServerHello(SslMessageBase):
 
 
 class TlsHandshakeMessageVariant(VariantParsable):
-    _VARIANTS = {
-        TlsHandshakeType.CLIENT_HELLO: TlsHandshakeClientHello,
-        TlsHandshakeType.SERVER_HELLO: TlsHandshakeServerHello,
-        TlsHandshakeType.CERTIFICATE: TlsHandshakeCertificate,
-        TlsHandshakeType.SERVER_KEY_EXCHANGE: TlsHandshakeServerKeyExchange,
-        TlsHandshakeType.SERVER_HELLO_DONE: TlsHandshakeServerHelloDone,
-    }
+    _VARIANTS = collections.OrderedDict([
+        (TlsHandshakeType.CLIENT_HELLO, [TlsHandshakeClientHello, ]),
+        (TlsHandshakeType.SERVER_HELLO, [TlsHandshakeServerHello, ]),
+        (TlsHandshakeType.CERTIFICATE, [TlsHandshakeCertificate, ]),
+        (TlsHandshakeType.SERVER_KEY_EXCHANGE, [TlsHandshakeServerKeyExchange, ]),
+        (TlsHandshakeType.CERTIFICATE_REQUEST, [TlsHandshakeCertificateRequest, ]),
+        (TlsHandshakeType.SERVER_HELLO_DONE, [TlsHandshakeServerHelloDone, ]),
+    ])
 
     @classmethod
     def _get_variants(cls):
